//
// Random Program Generator
// Bryan Turner (bryan.turner@pobox.com)
// July, 2005
//

#include "Common.h"
#include "Type.h"
#include "Function.h"
#include "Block.h"
#include "Variable.h"
#include "Statement.h"

vector<Function*> FuncList;		// List of all functions in the program
int cur_func_idx;				// Index into FuncList that we are currently working on
static bool param_first=true;					// Flag to track output of commas

unsigned long FuncListSize()
{
	return FuncList.size();
}

Function *GetFirstFunction()
{
	return FuncList[0];
}

string RandomFunctionName()
{
    return string( "func_" + RandomDigits(8) );
}

Type RandomReturnType()
{
	// TODO: Select a random type that is valid as a return type.
	Type t;
	t.eType = eSimple;
	t.simple_type = eInt;
	return t;
}

Function *SelectFunction( Function &curFunc, bool &isBackLink )
{
	// Select between creating a new function, or calling an old function..
	if ( FuncList.empty() || ((genrand()%1000) > 500) )
	{
		isBackLink = false;
		return new Function( curFunc );
	}
	else
	{
		// Link to a previously defined function (this may cause cycles in the call graph!)
		int idx = genrand() % FuncList.size();
		isBackLink = (idx <= cur_func_idx);
		return FuncList[idx];
	}
}

void Function::Init()
{
	STMT_DEPTH=0;
	body = 0;
	flags = 0;
	
	FuncList.push_back( this );				// Add to global list of functions..
	name = RandomFunctionName();
	return_type = RandomReturnType();
}

Function::Function( )
{
	Init();
	// No Parameter List
}

bool ParamListProbability( vector<Variable*> &List )
{
	if ( List.empty() ) return true;
	return (genrand() % 1000) > 500;		// 50% probability
}

void GenerateParameterList( Function &curFunc, Function &parent )
{
	while ( ParamListProbability( curFunc.param ) )
	{
		// With some probability, choose a new random variable, or one from parentParams,
		//      parentLocals, or the globals list...
		// Also, build the parent's link structure for this invocation and push it onto the back.
		GenerateParameterVariable( curFunc, parent );
	}
}

Function::Function( Function &parent )
{
	Init();
	GenerateParameterList( *this, parent );	// Select a set of vars to operate on..
}

void OutputFormalParam( Variable *var, std::ostream *pOut )
{
	std::ostream &out = *pOut;
	if ( !param_first ) out << ", ";
	param_first = false;
	var->type->Output( out );
	out << " " << var->name;
}

void Function::OutputFormalParamList( std::ostream &out )
{
	param_first = true;
	for_each( param.begin(), param.end(), std::bind2nd(std::ptr_fun(OutputFormalParam), &out) );
}

void Function::OutputHeader( std::ostream &out )
{
	return_type.Output( out );
	out << " " << name << "( ";
	OutputFormalParamList( out );
	out << " )";
}

void Function::OutputForwardDecl( std::ostream &out )
{
	OutputHeader( out );
	out << ";" << eol;
}

void Function::Output( std::ostream &out )
{
	out << "/* ------------------------------------------ */" << eol;
	OutputHeader( out );
	out << eol;
	out << "{" << eol;
	out << "if (DEPTH < MAX_DEPTH) " << eol;
	body->Output( out );
	out << "else" << eol;
	
	// TODO: Needs to be fixed when return types are no longer simple types..
	out << "return " << GenerateRandomSimpleConstant( return_type.simple_type ) << ";" << eol;
	out << "}" << eol;
	out << eol << eol;
};

void GenerateFunctionBody( Function *func )
{
	if ( !func->body )
	{
		// Fill in the Function body.
		func->body = new Block( *func, true );

// TODO: Re-enable this when the Statement class is separated into appropriate sub-classes.
		// Make sure there's an appropriate return statement (if required)..
		if ( func->return_type.eType != eSimple ||
			func->return_type.simple_type != eVoid )
		{
			func->stack.push_back( func->body );
			func->body->stm.push_back( new Statement( *func, eReturn ) );
			func->stack.pop_back();
		}

	}
}

void GenerateFunctions()
{
	// -----------------
	// Create a basic first function, then generate a random graph from there.
	Function *first = new Function();
	
	// -----------------
	// Create body of each function, continue until no new functions are created.
	for ( cur_func_idx = 0; cur_func_idx < FuncListSize(); cur_func_idx++ )
	{
		// Dynamically adds new functions to the end of the list..
		GenerateFunctionBody( FuncList[cur_func_idx] );
	}
}

void OutputForwardDecl( Function *func, std::ostream *pOut )
{
	func->OutputForwardDecl( *pOut );
}

void OutputFunction( Function *func, std::ostream *pOut )
{
	func->Output( *pOut );
}

void OutputForwardDeclarations( std::ostream &out )
{
	out << eol << eol << "/* --- FORWARD DECLARATIONS --- */" << eol;
	for_each( FuncList.begin(), FuncList.end(), std::bind2nd(std::ptr_fun(OutputForwardDecl), &out) );
}

void OutputFunctions( std::ostream &out )
{
	out << eol << eol << "/* --- FUNCTIONS --- */" << eol;
	for_each( FuncList.begin(), FuncList.end(), std::bind2nd(std::ptr_fun(OutputFunction), &out) );
}




