//
// Random Program Generator
// Bryan Turner (bryan.turner@pobox.com)
// July, 2005
//
// Contributions and bug fixes by:
// Jan-2007 : Mat Hostetter - Explicit "return 0" for main()
//

/*
Random C/C++ Program Generator
------------------------------
1) Create a set of random types to be used throughout the program
2) Create the main func.
3) Generate a random block of statements with maximum control & expression nesting depths.
4) If new functions were defined in #3, then loop back to fill in it's body.
5) When a maximum number of functions is reached, stop generating new functions and finish off the bodies of the remaining funcs.
6) Output generated program.

GOALS:
- Locate basic errors in compiler front-ends (crashes, etc).
- Ferret out correctness errors in optimization passes.
- Support the design of tools to search for improved optimization paths (partial exection, etc).

TODO:
- Protect back links with a global DEPTH, don't call if DEPTH is too high (avoid infinite recursion)
- Main function generates hash of global state as output of program - use to locate optimization errors.
	- Compile in Debug mode vs. Optimized mode, compare hash value at program termination.
- Improve hash function; use stronger hashing (ARCFOUR) and perform hashing at random points in the graph.
	- Output only after successful program termination.

FUTURE:
- Complex types
- Type-correct expressions
- Some substitutions allowed
	- int, char, short, long, float, double - all interchangeable to some degree (use appropriate casts)
	- array <--> pointer
	- pointer promotion (ie: passing the pointer to a local var, or droping the pointer to pass by value)
- Memory Allocation & Manipulation?
	- Choose random functions to perform allocations
	- Choose random children/ancestors to perform deallocations
	- Work from leaves to root
	- If node uses pointer or array, it is potential heap store allocated. 
*/

#include "Common.h"
#include "Type.h"
#include "Block.h"
#include "Function.h"
#include "Variable.h"
#include "Statement.h"
#include "Expression.h"
#include "FunctionInvocation.h"

// ------------------------------------------------------------------------------------------
// Globals
const string eol("\n");
bool g_first;					// Flag to track output of commas
unsigned long g_Seed = 0;		// Program seed - allow user to regenerate the same program on different platforms.

// ------------------------------------------------------------------------------------------
// A C-program for TT800 : July 8th 1996 Version
// by M. Matsumoto, email: matumoto@math.keio.ac.jp
// genrand() generate one pseudorandom number with double precision
// which is uniformly distributed on [0,1]-interval
// for each call.  One may choose any initial 25 seeds
// except all zeros.
// http://random.mat.sbg.ac.at/ftp/pub/data/tt800.c

/* See: ACM Transactions on Modelling and Computer Simulation, */
/* Vol. 4, No. 3, 1994, pages 254-266. */

#include <stdio.h>
#define N 25
#define M 7

/* initial 25 seeds, change as you wish */
static unsigned long x[N] =
{
	0x95f24dab, 0x0b685215, 0xe76ccae7, 0xaf3ec239, 0x715fad23,
	0x24a590ad, 0x69e4b5ef, 0xbf456141, 0x96bc1b7b, 0xa7bdf825,
	0xc1de75b7, 0x8858a9c9, 0x2da87693, 0xb657f9dd, 0xffdc8a9f,
	0x8121da71, 0x8b823ecb, 0x885d05f5, 0x4e20cd47, 0x5a9ad5d9,
	0x512c0c03, 0xea857ccd, 0x4cc1d30f, 0x8891a8a1, 0xa6b7aadb
};

void seedrand( unsigned long seed )
{
	for (int kk=0;kk<N;kk++)
	{
		x[kk] ^= seed;
	}
}

unsigned long genrand()
{
	unsigned long y;
	static int k = 0;
	static unsigned long mag01[2] =
	{ 
		0x0, 0x8ebfd028 /* this is magic vector `a', don't change */
	};
	
 	/* generate N words at one time */
 	if ( k == N )
	{
		int kk;
		for (kk=0;kk<N-M;kk++)
		{
			x[kk] = x[kk+M] ^ (x[kk] >> 1) ^ mag01[x[kk] % 2];
		}
		
		for (; kk<N;kk++)
		{
			x[kk] = x[kk+(M-N)] ^ (x[kk] >> 1) ^ mag01[x[kk] % 2];
		}
		k=0;
	}
	y = x[k];
	y ^= (y << 7) & 0x2b5b2500; /* s and b, magic vectors */
	y ^= (y << 15) & 0xdb8b0000; /* t and c, magic vectors */
	y &= 0xffffffff; /* you may delete this line if word size = 32 */
	/*
	the following line was added by Makoto Matsumoto in the 1996 version
	to improve lower bit's corellation.
	Delete this line to use the code published in 1994.
	*/
	y ^= (y >> 16); /* added to the 1994 version */
	k++;
	return  y;
}

// ------------------------------------------------------------------------------------------
const char hex[] = "0123456789ABCDEF";
const char dec[] = "0123456789";

string RandomHexDigits( int num )
{
	string str;
	while ( num-- )
	{
		str += hex[genrand()%16];
	}

	return str;
}

string RandomDigits( int num )
{
	string str;
	while ( num-- )
	{
		str += dec[genrand()%10];
	}

	return str;
}

// --------------------------------------------------------------------------------
void OutputHeader( std::ostream &out )
{
	// Should output the seed info, generator version, statistical sampling, etc..
	out << "/*" << eol << "// --- RANDOMLY GENERATED PROGRAM ---" << eol;
	out << "// Program Generator by Bryan Turner (bryan.turner@pobox.com)" << eol;
	out << "//" << eol;
	out << "// Seed: " << g_Seed << eol;
	out <<"*/" << eol << eol;
	out << "#include <stdio.h>" << eol;
	out << eol;
	out << "#define MAX_DEPTH (5)" << eol;
	out << "unsigned long context = 0;" << eol;
	out << "long DEPTH = 0;" << eol;		// Make depth signed, to cover our tails..
//	out << "void DumbHash( unsigned char *value, unsigned int len )" << eol;
//	out << "{\nwhile ( len-- )\n{\ncontext += *(value++);\n}\ncontext ^= 0xA50F5AF0;\n}" << eol;
	out << "void DumbHash( unsigned long value, unsigned int len )" << eol;
	out << "{" << eol;
	out << "   context += value;" << eol;
	out << "   context ^= 0xA50F5AF0;" << eol;
//	out << "   printf( \"%d\\n\", context );" << eol
	out << "}" << eol;
}

// --------------------------------------------------------------------------------
void OutputMain( std::ostream &out )
{
	FunctionInvocation *invoke = new FunctionInvocation( GetFirstFunction() );
	out << eol << eol << "/* ---------------------------------------- */" << eol;
	out << "int main( int argc, void *argv[] )" << eol << "{" << eol;
	out << "   /* Call the first function */" << eol;
	out << "   ";
	invoke->Output( out );
	out << ";" << eol;
	HashGlobalVariables( out );
	out << "printf( \"%d\\n\", context );" << eol;
	out << "return 0;" << eol;
	out << "}" << eol << eol << eol;
}

// --------------------------------------------------------------------------------
int main()
{
	g_Seed = platform_gen_seed();
	seedrand( g_Seed );

	// Create a list of types that will be used by this program
	GenerateAllTypes();
	GenerateFunctions();

	// Output the random program..
//	std::ofstream file;			// Output file
//	file.open( "junk.cpp" );
//	std::ostream &out = file;

	std::ostream &out = std::cout;
	OutputHeader( out );
	OutputGlobalVariables( out );
	OutputForwardDeclarations( out );
	OutputFunctions( out );
	OutputMain( out );
	
//	file.close();
}




