//
// Random Program Generator
// Bryan Turner (bryan.turner@pobox.com)
// July, 2005
//

#include "Common.h"
#include "Type.h"
#include "Expression.h"
#include "Variable.h"
#include "Block.h"
#include "Function.h"
#include "FunctionInvocation.h"
#include "Statement.h"

// ------------------------------------------------------------------------------------------
//
// TODO: Build probability tables for each decision point
//       IE: Return statements less common than assignment statements, etc.
//       Allow MIN, MAX and pluggable density functions (uniform, normal, guassian, ...)
//
// Like this:
/*
unsigned int probabilityStatement[eStatementType] =
{
	100, 	//	eAssign,
	75, 	//	eIfElse,
	30,		//	eFor,		// Make this a generic loop construct (while/for/do)
	30,		//	eInvoke,
	10,		//	eReturn,
}

probability probStatement
{
	2, // Min
	0, // Max
	probabilityStatement // Pluggable Function (custom probability table, in this case)
}
*/

eStatementType StatementProbability( Function &parent )
{
	eStatementType eType;
	while (1)
	{
		eType = (eStatementType) (genrand() % MAX_STATEMENT_TYPE);

//		if ( eType == eHash ) continue;

		if ( (eType == eReturn) && (parent.flags & NO_RETURN) )
			continue;

// TODO: Re-enable continue/break when LLVM bug is fixed..
//		if ( (eType == eBreak || eType == eContinue )
//              && !(parent.flags & IN_LOOP) )
//			continue;

		// Limit Function complexity (depth of nested control structures and expressions)
		if ( parent.STMT_DEPTH > MAX_STMT_DEPTH )
		{
			if ( eType == eAssign || eType == eReturn )
				return eType;
		}
		else if ( FuncListSize() > MAX_FUNCTIONS )		// Limit # of functions..
		{
			if ( eType != eInvoke )
				return eType;
		}
		else
            return eType;
	}
}

void Statement::IfElse( Function &curFunc )
{
	eType = eIfElse;
	expr = new Expression( curFunc );
	if_true = new Block( curFunc );
	if_false = new Statement( curFunc );
}

void Statement::Invoke( Function &curFunc )
{
	eType = eInvoke;
	invoke = new FunctionInvocation( curFunc, false );
}

void Statement::Return( Function &curFunc )
{
	eType = eReturn;

	if ( curFunc.return_type.eType != eSimple ||
		 curFunc.return_type.simple_type != eVoid )
	{
		var = SelectVariable( curFunc ); //SelectVariableOfType( curFunc.return_type );
	}
}

void Statement::Assign( Function &curFunc )
{
	eType = eAssign;
	var = SelectLValue( curFunc );
	expr = new Expression( curFunc );
}

void Statement::For( Function &curFunc )
{
	// Trivial FOR statement - for ( int i=0; i<100; i++ )
	// This is generated in the OutputForStatement() function...
	eType = eFor;
	var = SelectLValue( curFunc );

	unsigned long flags_save = curFunc.flags;
	curFunc.flags |= IN_LOOP;
	if_true = new Block( curFunc );
	curFunc.flags = flags_save;
}

Statement::Statement( Function &curFunc, eStatementType t )
{
	// Add more statements:
	// for
	// while
	// library call (malloc, free, str*, mem*, etc..)?
	// switch?
	// ..?
	if ( t == MAX_STATEMENT_TYPE )
	{
		t = StatementProbability( curFunc );
	}
	
	switch ( t )
	{
	case eAssign: Assign( curFunc ); break;
	case eIfElse: IfElse( curFunc ); break;
	case eFor:    For( curFunc );	break;
	case eInvoke: Invoke( curFunc ); break;
	case eReturn: Return( curFunc ); break;
//	case eBreak:
//	case eContinue: stm = new Statement(); stm->eType = t;	break;
	}
}

void Statement::OutputReturn( std::ostream &out )
{
	out << "DEPTH--;" << eol;
	out << "return ";
	var->Output( out );
	out << ";" << eol;
}

void Statement::OutputIfElse( std::ostream &out )
{
	out << "if ( ";
	expr->Output( out );
	out << " )" << eol;
	if_true->Output( out );
	out << "else" << eol;
	if_false->Output( out );
}

void Statement::OutputAssign( std::ostream &out )
{
	var->Output( out );
	out << " = ";
	expr->Output( out );
	out << ";" << eol;
}

void Statement::OutputFor( std::ostream &out )
{
	// Trivial For loop..
	out << "for ( ";
	var->Output( out );
	out << " = 0; ";
	var->Output( out );
	out << " < 100; ++";
	var->Output( out );
	out << " )" << eol;
	if_true->Output( out );
}

void Statement::OutputInvoke( std::ostream &out )
{
// TODO: Need some way for this to work in the middle of an expression.
//       Or perhaps, all functions should have depth checks as their first operation?

//	if ( invoke->isBackLink )
//	{
//		out << "if ( DEPTH < 5 ) { ";
//	}

	invoke->Output( out );
	out << ";";

//	if ( invoke->isBackLink )
//	{
//		out << " }";
//	}
	
	out << eol;
}

void Statement::Output( std::ostream &out )
{
	switch (eType)
	{
	case eAssign: OutputAssign( out ); break;
	case eIfElse: OutputIfElse( out ); break;
	case eInvoke: OutputInvoke( out ); break;
	case eReturn: OutputReturn( out ); break;
//	case eBreak:  out << "break;" << eol; break;
//	case eContinue:  out << "continue;" << eol; break;
	case eFor: OutputFor( out ); break;
//	case eHash: OutputHashStatement( this ); break;
	}
};

