//
// Random Program Generator
// Bryan Turner (bryan.turner@pobox.com)
// July, 2005
//

#include "Common.h"
#include "Variable.h"
#include "Type.h"
#include "Block.h"
#include "Function.h"

// --------------------------------------------------------------
// Globals..
vector<Variable*> GlobalList;	// List of all globals in the program

// --------------------------------------------------------------
Variable *SelectConstant()
{
	// Select a random type and generate a constant for it.
	// .. make the type a Simple Type..
	Variable *var = new Variable();
	var->isConst = true;
	var->type = new Type();
	var->type->eType = eSimple;
	var->type->simple_type = (eSimpleType) (1+ (genrand() % (MAX_SIMPLE_TYPES-1)));	// Don't allow void constants..
	var->name = GenerateRandomSimpleConstant( var->type->simple_type );
	return var;
}

// --------------------------------------------------------------
string RandomGlobalName()
{
    return string( "g_" + RandomDigits(8) );
}

// --------------------------------------------------------------
string RandomLocalName()
{
    return string( "l_" + RandomDigits(8) );
}

// --------------------------------------------------------------
string RandomParamName()
{
    return string( "p_" + RandomDigits(8) );
}

// --------------------------------------------------------------
Variable *GenerateNewGlobal()
{
	GlobalList.push_back( new Variable() );
	Variable *var = GlobalList.back();
	var->type = SelectRandomType();
	var->name = RandomGlobalName();
	return var;
}

// --------------------------------------------------------------
Variable *SelectGlobal()
{
	// Select a random global variable..
	if ( GlobalList.empty() )
	{
		return GenerateNewGlobal();
	}
	return GlobalList[genrand() % GlobalList.size()];
}

// --------------------------------------------------------------
Variable *GenerateNewParentLocal( Block &block )
{
	block.local_vars.push_back( new Variable() );
    Variable *var = block.local_vars.back();
	var->type = SelectRandomType();
	var->name = RandomLocalName();
	return var;
}

// --------------------------------------------------------------
void GenerateParameterVariable( Function &curFunc, Function &parent )
{
	Variable *var = SelectVariable( parent );

	// Add this type to our parameter list..
	curFunc.param.push_back( new Variable() );		// This is the formal parameter for the function.
	curFunc.param.back()->name = RandomParamName();
	curFunc.param.back()->type = var->type;
}

// --------------------------------------------------------------
Variable *SelectParentLocal( Function &parent )
{
	// Select from the local variables of the parent OR any of it's block stack...
	//
	Block *block = parent.stack[genrand() % parent.stack.size()];
	
	// Should be "generate new block local"...
	if ( block->local_vars.empty() ) return GenerateNewParentLocal( *block );
	return block->local_vars[genrand() % block->local_vars.size()];
}

// --------------------------------------------------------------
eVariableScope VariableSelectionProbability( vector<Variable*> &List )
{
	// Should probably modify choice based on current list of params, parent params,
	//     parent locals, # of globals, etc..
	// .. but this is easier.
	return (eVariableScope)(genrand() % MAX_VAR_SCOPE);
}

// --------------------------------------------------------------
eVariableScope LValueSelectionProbability( vector<Variable*> &List )
{
	eVariableScope eType;
	while (1)
	{
		eType = VariableSelectionProbability( List );
		if ( eType != eConstant && eType != eNewValue )
			return eType;
	}
}

// --------------------------------------------------------------
eVariableScope VariableCreationProbability( Function &parent )
{
	if ( (genrand() % 1000) > 900 )		// 10% chance to create new global var
		return eGlobal;
	else
		return eParentLocal;
}

// --------------------------------------------------------------
Variable *SelectParentParam( Function &parent )
{
	if ( parent.param.empty() ) return SelectParentLocal( parent );	// Error case.. no params in parent. 
	return parent.param[genrand() % parent.param.size()];
}

// --------------------------------------------------------------
Variable *GenerateNewVariable( Function &func )
{
	Variable *var;
	switch ( VariableCreationProbability( func ) )
	{
	case eGlobal:		var = GenerateNewGlobal();	 break;					// Create a new global variable
	case eParentLocal:	var = GenerateNewParentLocal( *(func.stack[genrand() % func.stack.size()]) );	break;		// Create a new local variable in the parent
	}

	return var;
}

// --------------------------------------------------------------
Variable *SelectLValue( Function &func )
{
	Variable *var;
	switch ( LValueSelectionProbability( func.param ) )
	{
	case eGlobal:		var = SelectGlobal(); break;					// Select a global
	case eParentLocal:	var = SelectParentLocal( func ); break;			// .. a local var from one of it's blocks..
	case eParentParam:	var = SelectParentParam( func ); break;			// .. one of the function's parameters
	}
	return var;
}

// --------------------------------------------------------------
// Select or Create a new variable visible to this scope (new var may be global,
//    or local to one of the function's blocks)
Variable *SelectVariable( Function &func )
{
	Variable *var;
	switch ( VariableSelectionProbability( func.param ) )
	{
	case eConstant:		var = SelectConstant(); break;					// Select a constant value of some type..
	case eGlobal:		var = SelectGlobal(); break;					// Select a global
	case eParentLocal:	var = SelectParentLocal( func ); break;			// .. a local var from one of it's blocks..
	case eParentParam:	var = SelectParentParam( func ); break;			// .. one of the function's parameters
	case eNewValue:		var = GenerateNewVariable( func ); break;		// Must decide where to put the new variable (global or parent local)?
	}
	return var;
}

// --------------------------------------------------------------
Variable *SelectExistingVariable( Function &func )
{
	Variable *var = 0;
	while ( !var )
	{
		switch ( VariableSelectionProbability( func.param ) )
		{
		case eGlobal:		var = SelectGlobal(); break;			// Select a global
		case eParentLocal:	var = SelectParentLocal( func ); break;	// .. a local var from one of it's blocks..
		case eParentParam:	var = SelectParentParam( func ); break;	// .. one of the function's parameters
		}
	}
	return var;
}

// --------------------------------------------------------------
void Variable::OutputDef( std::ostream &out ) const
{
	type->Output( out );
	out << " " << name << " = " << GenerateRandomSimpleConstant( type->simple_type ) << ";" << eol;
};

// --------------------------------------------------------------
void Variable::Output( std::ostream &out ) const
{
	out << name;
};

// --------------------------------------------------------------
void MapVariableList( vector<Variable*> &var, std::ostream &out, void (*func)(Variable *var, std::ostream *pOut) )
{
	for_each( var.begin(), var.end(), std::bind2nd(std::ptr_fun(func), &out) );
}

// --------------------------------------------------------------
void OutputVariable( Variable *var, std::ostream *pOut )
{
	var->OutputDef( *pOut );
}

// --------------------------------------------------------------
void OutputVariableList( vector<Variable*> &var, std::ostream &out )
{
	MapVariableList( var, out, OutputVariable );
}

// --------------------------------------------------------------
void OutputGlobalVariables( std::ostream &out )
{
	out << "/* --- GLOBAL VARIABLES --- */" << eol;
	OutputVariableList( GlobalList, out );
}

// --------------------------------------------------------------
void HashVariable( Variable *var, std::ostream *pOut )
{
	std::ostream &out = *pOut;
	out << "DumbHash( ";
	var->Output( out );
	out << ", " << var->type->SizeInBytes() << " );" << eol;
}

// --------------------------------------------------------------
void HashGlobalVariables( std::ostream &out )
{
	MapVariableList( GlobalList, out, HashVariable );
}


